import pytest

from .queues import ProfilingQueue, Deque, QueuePolicy

all_queues = [
    (ProfilingQueue, ()),
    (Deque, ()),
    (Deque, (QueuePolicy.DEPTHFIRST,)),
    (Deque, (QueuePolicy.BREADTHFIRST,)),
]


@pytest.mark.parametrize(
    "queue,queue_args", all_queues,
)
def test_empty(queue, queue_args):
    q = queue(*queue_args)
    assert q.empty()
    q.put(1)
    assert q.empty() == False
    q.get()
    assert q.empty()


@pytest.mark.parametrize(
    "queue,queue_args", all_queues,
)
def test_can_put_and_pop(queue, queue_args):
    q = queue(*queue_args)
    q.put(1)
    assert q.get() == 1
    q.put(2)
    q.put(3)
    q.put(lambda x: x)

    q.get()
    q.get()
    q.get()
    assert q.empty()


def test_bfs_dfs():
    bfs = Deque(QueuePolicy.BREADTHFIRST)
    dfs = Deque(QueuePolicy.DEPTHFIRST)

    bfs.put(1)
    dfs.put(1)
    bfs.put(2)
    dfs.put(2)

    assert bfs.get() == 2
    assert dfs.get() == 1



"""
In this process, we consider the following operations to be fast:

- Get degree of node
- Get downstream targets of node
- Get attributes on a node

These operations are medium:

- Get upstream sources of node
- Get degrees of all nodes in host graph
- Get nodes with a certain attribute

These operations are slow:
- Get edges where nodes have a certain attribute
    - But you can do the same with get-downstream-targets and filter an
      attribute search.
"""

from typing import Dict, Generator, Hashable, List, Optional, Union, Tuple
from inspect import isclass
import itertools
from functools import lru_cache

import networkx as nx
from .queues import SimpleQueue

__version__ = "2.1.1"


@lru_cache()
def _is_node_attr_match(
    motif_node_id: str, host_node_id: str, motif: nx.Graph, host: nx.Graph
) -> bool:
    """
    Check if a node in the host graph matches the attributes in the motif.

    Arguments:
        motif_node_id (str): The motif node ID
        host_node_id (str): The host graph ID
        motif (nx.Graph): The motif graph
        host (nx.Graph): The host graph

    Returns:
        bool: True if the host node matches the attributes in the motif

    """
    motif_node = motif.nodes[motif_node_id]
    host_node = host.nodes[host_node_id]

    for attr, val in motif_node.items():
        if attr not in host_node:
            return False
        if host_node[attr] != val:
            return False

    return True


@lru_cache()
def _is_node_structural_match(
    motif_node_id: str, host_node_id: str, motif: nx.Graph, host: nx.Graph
) -> bool:
    """
    Check if the motif node here is a valid structural match.

    Specifically, this requires that a host node has at least the degree as the
    motif node.

    Arguments:
        motif_node_id (str): The motif node ID
        host_node_id (str): The host graph ID
        motif (nx.Graph): The motif graph
        host (nx.Graph): The host graph

    Returns:
        bool: True if the motif node maps to this host node

    """
    return host.degree(host_node_id) >= motif.degree(motif_node_id)


@lru_cache()
def _is_edge_attr_match(
    motif_edge_id: Tuple[str, str],
    host_edge_id: Tuple[str, str],
    motif: nx.Graph,
    host: nx.Graph,
) -> bool:
    """
    Check if an edge in the host graph matches the attributes in the motif.

    Arguments:
        motif_edge_id (str): The motif edge ID
        host_edge_id (str): The host edge ID
        motif (nx.Graph): The motif graph
        host (nx.Graph): The host graph

    Returns:
        bool: True if the host edge matches the attributes in the motif

    """
    motif_edge = motif.edges[motif_edge_id]
    host_edge = host.edges[host_edge_id]

    for attr, val in motif_edge.items():
        if attr not in host_edge:
            return False
        if host_edge[attr] != val:
            return False

    return True


def get_next_backbone_candidates(
    backbone: dict,
    motif: nx.Graph,
    host: nx.Graph,
    interestingness: dict,
    next_node: str = None,
    directed: bool = True,
    is_node_structural_match=_is_node_structural_match,
    is_node_attr_match=_is_node_attr_match,
    is_edge_attr_match=_is_edge_attr_match,
    isomorphisms_only: bool = False,
) -> List[dict]:
    """
    Get a list of candidate node assignments for the next "step" of this map.

    Arguments:
        backbone (dict): Mapping of motif node IDs to one set of host graph IDs
        motif (Graph): A graph representation of the motif
        host (Graph): The host graph, complete
        interestingness (dict): A mapping of motif node IDs to interestingness
        next_node (str: None): Optional suggestion for the next node to assign
        directed (bool: True): Whether host and motif are both directed
        isomorphisms_only (bool: False): If true, only isomorphisms will be
            returned (instead of all monomorphisms)

    Returns:
        List[dict]: A new list of mappings with one additional element mapped

    """

    # Get a list of the "exploration front" of the motif -- nodes that are not
    # yet assigned in the backbone but are connected to at least one assigned
    # node in the backbone.

    # For example, in the motif A -> B -> C, if A is already assigned, then the
    # front is [B] (c is not included because it has not connection to any
    # assigned node).

    # We should prefer nodes that are connected to multiple assigned backbone
    # nodes, because these will filter more rapidly to a smaller set.

    # First check if the backbone is empty. If so, we should choose the most
    # interesting node to start with:
    if next_node is None and len(backbone) == 0:
        # This is the starting-case, where we have NO backbone nodes set yet.
        next_node = max(
            interestingness.keys(), key=lambda node: interestingness.get(node, 0.0)
        )
        # Let's return ALL possible node choices for this next_node. To do this
        # without being an insane person, let's filter on max degree in host:
        return [
            {next_node: n}
            for n in host.nodes
            if is_node_attr_match(next_node, n, motif, host)
            and is_node_structural_match(next_node, n, motif, host)
        ]

    else:
        _nodes_with_greatest_backbone_count: List[str] = []
        _greatest_backbone_count = 0
        for motif_node_id in motif.nodes:
            if motif_node_id in backbone:
                continue
            # How many connections to existing backbone?
            # Note that this number is certainly greater than or equal to 1,
            # since a value of 0 would imply that the backbone dict is empty
            # (which we have already handled) or that the motif has more than
            # one connected component, which we check for at prep-time.
            if directed:
                motif_backbone_connections_count = sum(
                    [
                        1
                        for v in list(
                            set(motif.adj[motif_node_id]).union(
                                set(motif.pred[motif_node_id])
                            )
                        )
                        if v in backbone
                    ]
                )
            else:
                motif_backbone_connections_count = sum(
                    [1 for v in motif.adj[motif_node_id] if v in backbone]
                )
            # If this is the most highly connected node visited so far, then
            # set it as the next node to explore:
            if motif_backbone_connections_count > _greatest_backbone_count:
                _nodes_with_greatest_backbone_count.append(motif_node_id)
        # Now we have _node_with_greatest_backbone_count as the best candidate
        # for `next_node`.
        next_node = max(
            _nodes_with_greatest_backbone_count,
            key=lambda node: interestingness.get(node, 0.0),
        )

    # Now we have a node `next_node` which we know is connected to the current
    # backbone. Get all edges between `next_node` and nodes in the backbone,
    # and verify that they exist in the host graph:
    # `required_edges` has the form (prev, self, next), with non-values filled
    # with None. That way we can easily remember and store the roles of the
    # node IDs in the next step.
    required_edges = []
    for other in list(motif.adj[next_node]):
        if other in backbone:
            # edge is (next_node, other)
            required_edges.append((None, next_node, other))
    if directed:
        for other in list(motif.pred[next_node]):
            if other in backbone:
                # edge is (other, next_node)
                required_edges.append((other, next_node, None))

    # `required_edges` now contains a list of all edges that exist in the motif
    # graph, and we must find candidate nodes that have such edges in the host.

    candidate_nodes = []

    # In the worst-case, `required_edges` has length == 1. This is the worst
    # case because it means that ALL edges from/to `other` are valid options.
    if len(required_edges) == 1:
        # :(
        (source, _, target) = required_edges[0]
        if directed:
            if source is not None:
                # this is a "from" edge:
                candidate_nodes = list(host.adj[backbone[source]])
            elif target is not None:
                # this is a "from" edge:
                candidate_nodes = list(host.pred[backbone[target]])
        else:
            candidate_nodes = list(host.adj[backbone[target]])
        # Thus, all candidates for motif ID `$next_node` are stored in the
        # candidate_nodes list.

    elif len(required_edges) > 1:
        # This is neato :) It means that there are multiple edges in the host
        # graph that we can use to downselect the number of candidate nodes.
        candidate_nodes_set = set()
        for (source, _, target) in required_edges:
            if directed:
                if source is not None:
                    # this is a "from" edge:
                    candidate_nodes_from_this_edge = host.adj[backbone[source]]
                # elif target is not None:
                else:  # target is not None:
                    # this is a "from" edge:
                    candidate_nodes_from_this_edge = host.pred[backbone[target]]
                # else:
                #     raise AssertionError("Encountered an impossible condition: At least one of source or target must be defined.")
            else:
                candidate_nodes_from_this_edge = host.adj[backbone[target]]
            if len(candidate_nodes_set) == 0:
                # This is the first edge we're checking, so set the candidate
                # nodes set to ALL possible candidates.
                candidate_nodes_set.update(candidate_nodes_from_this_edge)
            else:
                candidate_nodes_set = candidate_nodes_set.intersection(
                    candidate_nodes_from_this_edge
                )
        candidate_nodes = list(candidate_nodes_set)

    elif len(required_edges) == 0:
        # Somehow you found a node that doesn't have any edges. This is bad.
        raise ValueError(
            f"Somehow you found a motif node {next_node} that doesn't have "
            + "any motif-graph edges. This is bad. (Did you maybe pass an "
            + "empty backbone to this function?)"
        )

    tentative_results = [
        {**backbone, next_node: c}
        for c in candidate_nodes
        if c not in backbone.values()
        and is_node_attr_match(next_node, c, motif, host)
        and is_node_structural_match(next_node, c, motif, host)
    ]

    # One last filtering step here. This is to catch the cases where you have
    # successfully mapped each node, and the final node has some valid
    # candidate_nodes (and therefore `tentative_results`).
    # This is important: We must now check that for the assigned nodes, all
    # edges between them DO exist in the host graph. Otherwise, when we check
    # in find_motifs that len(motif) == len(mapping), we will discover that the
    # mapping is "complete" even though we haven't yet checked it at all.

    monomorphism_candidates = []

    for mapping in tentative_results:
        if len(mapping) == len(motif):
            if all(
                [
                    host.has_edge(mapping[motif_u], mapping[motif_v])
                    and is_edge_attr_match(
                        (motif_u, motif_v),
                        (mapping[motif_u], mapping[motif_v]),
                        motif,
                        host,
                    )
                    for motif_u, motif_v in motif.edges
                ]
            ):
                # This is a "complete" match!
                monomorphism_candidates.append(mapping)
        else:
            # This is a partial match, so we'll continue building.
            monomorphism_candidates.append(mapping)

    if not isomorphisms_only:
        return monomorphism_candidates

    # Additionally, if isomorphisms_only == True, we can use this opportunity
    # to confirm that no spurious edges exist in the induced subgraph:
    isomorphism_candidates = []
    for result in monomorphism_candidates:
        for (motif_u, motif_v) in itertools.product(result.keys(), result.keys()):
            # if the motif has this edge, then it doesn't rule any of the
            # above results out as an isomorphism.
            # if the motif does NOT have the edge, then NO RESULT may have
            # the equivalent edge in the host graph:
            if not motif.has_edge(motif_u, motif_v) and host.has_edge(
                result[motif_u], result[motif_v]
            ):
                # this is a violation.
                break
        else:
            isomorphism_candidates.append(result)
    return isomorphism_candidates


def uniform_node_interestingness(motif: nx.Graph) -> dict:
    """
    Sort the nodes in a motif by their interestingness.

    Most interesting nodes are defined to be those that most rapidly filter the
    list of nodes down to a smaller set.

    """
    return {n: 1 for n in motif.nodes}


def find_motifs_iter(
    motif: nx.Graph,
    host: nx.Graph,
    interestingness: dict = None,
    directed: bool = None,
    queue_=SimpleQueue,
    isomorphisms_only: bool = False,
    hints: List[Dict[Hashable, Hashable]] = None,
    is_node_structural_match=_is_node_structural_match,
    is_node_attr_match=_is_node_attr_match,
    is_edge_attr_match=_is_edge_attr_match,
) -> Generator[dict, None, None]:
    """
    Yield mappings from motif node IDs to host graph IDs.

    Results are of the form:

    ```
    {motif_id: host_id, ...}
    ```

    Arguments:
        motif (nx.DiGraph): The motif graph (needle) to search for
        host (nx.DiGraph): The host graph (haystack) to search within
        interestingness (dict: None): A map of each node in `motif` to a float
            number that indicates an ordinality in which to address each node
        directed (bool: None): Whether direction should be considered during
            search. If omitted, this will be based upon the motif directedness.
        queue_ (queue.SimpleQueue): What kind of queue to use.
        hints (dict): A dictionary of initial starting mappings. By default,
            searches for all instances. You can constrain a node by passing a
            list with a single dict item: `[{motifId: hostId}]`.
        isomorphisms_only (bool: False): Whether to return isomorphisms (the
            default is monomorphisms).

    Returns:
        Generator[dict, None, None]

    """
    interestingness = interestingness or uniform_node_interestingness(motif)
    if directed is None:
        # guess directedness from motif
        if isinstance(motif, nx.DiGraph):
            # This will be a directed query.
            directed = True
        else:
            directed = False

    q = queue_() if isclass(queue_) else queue_

    # Kick off the queue with an empty candidate:
    if hints is None or hints == []:
        q.put({})
    else:
        for hint in hints:
            q.put(hint)

    while not q.empty():
        new_backbone = q.get()
        next_candidate_backbones = get_next_backbone_candidates(
            new_backbone,
            motif,
            host,
            interestingness,
            directed=directed,
            isomorphisms_only=isomorphisms_only,
            is_node_structural_match=is_node_structural_match,
            is_node_attr_match=is_node_attr_match,
            is_edge_attr_match=is_edge_attr_match,
        )

        for candidate in next_candidate_backbones:
            if len(candidate) == len(motif):
                yield candidate
            else:
                q.put(candidate)


def find_motifs(
    motif: nx.Graph,
    host: nx.Graph,
    *args,
    count_only: bool = False,
    limit: int = None,
    is_node_attr_match=_is_node_attr_match,
    is_node_structural_match=_is_node_structural_match,
    is_edge_attr_match=_is_edge_attr_match,
    **kwargs,
) -> Union[int, List[dict]]:
    """
    Get a list of mappings from motif node IDs to host graph IDs.

    Results are of the form:

    ```
    [{motif_id: host_id, ...}]
    ```

    See grandiso#find_motifs_iter for full argument list.

    Arguments:
        count_only (bool: False): If True, return only an integer count of the
            number of motifs, rather than a list of mappings.
        limit (int: None): A limit to place on the number of returned mappings.
            The search will terminate once the limit is reached.


    Returns:
        int: If `count_only` is True, return the length of the List.
        List[dict]: A list of mappings from motif node IDs to host graph IDs

    """
    results = []
    results_count = 0
    for qresult in find_motifs_iter(
        motif,
        host,
        *args,
        is_node_attr_match=is_node_attr_match,
        is_node_structural_match=is_node_structural_match,
        is_edge_attr_match=is_edge_attr_match,
        **kwargs,
    ):

        result = qresult

        results_count += 1
        if limit and results_count >= limit:
            if count_only:
                return results_count
            else:
                # Subtract 1 from results_count because we have not yet
                # added the new result to the results list, but we HAVE
                # already added +1 to the count.
                if limit and (results_count - 1) >= limit:
                    return results
        if not count_only:
            results.append(result)

    if count_only:
        return results_count
    return results



import time
import copy
import random
import pytest

import networkx as nx
from networkx.algorithms.isomorphism import DiGraphMatcher, GraphMatcher

from . import find_motifs, find_motifs_iter


class TestSubgraphMatching:
    def test_finds_no_triangles_in_zero_tri_graph(self):

        motif = nx.DiGraph()
        motif.add_edge("A", "B")
        motif.add_edge("B", "C")
        motif.add_edge("C", "A")

        host = nx.DiGraph()
        host.add_edge("A", "B")
        host.add_edge("B", "C")

        assert len(list(find_motifs_iter(motif, host))) == 0

    def test_finds_no_rect_in_zero_rect_graph(self):

        motif = nx.DiGraph()
        motif.add_edge("A", "B")
        motif.add_edge("B", "C")
        motif.add_edge("C", "D")
        motif.add_edge("D", "A")

        host = nx.DiGraph()
        host.add_edge("A", "B")
        host.add_edge("B", "C")
        host.add_edge("C", "D")

        assert len(find_motifs(motif, host)) == 0

    def test_finds_no_triangles_in_zero_tri_graph_with_context(self):

        motif = nx.DiGraph()
        motif.add_edge("A", "B")
        motif.add_edge("B", "C")
        motif.add_edge("C", "A")
        motif.add_edge("C", "D")
        motif.add_edge("C", "E")

        host = nx.DiGraph()
        host.add_edge("A", "B")
        host.add_edge("B", "C")

        assert len(list(find_motifs_iter(motif, host))) == 0

    def test_finds_no_motifs_in_small_graph(self):

        motif = nx.DiGraph()
        motif.add_edge("A", "B")
        motif.add_edge("B", "C")

        host = nx.DiGraph()
        host.add_edge("A", "B")

        assert len(find_motifs(motif, host)) == 0

    def test_subgraph_equals_graph_triangle(self):

        motif = nx.DiGraph()
        motif.add_edge("A", "B")
        motif.add_edge("B", "C")
        motif.add_edge("C", "A")

        host = nx.DiGraph()
        host.add_edge("A", "B")
        host.add_edge("B", "C")
        host.add_edge("C", "A")

        assert len(find_motifs(motif, host)) == 3

    def test_subgraph_equals_graph_triangle_count_only(self):

        motif = nx.DiGraph()
        motif.add_edge("A", "B")
        motif.add_edge("B", "C")
        motif.add_edge("C", "A")

        host = nx.DiGraph()
        host.add_edge("A", "B")
        host.add_edge("B", "C")
        host.add_edge("C", "A")

        assert find_motifs(motif, host, count_only=True) == 3

    def test_subgraph_equals_graph_rect(self):

        motif = nx.DiGraph()
        motif.add_edge("A", "B")
        motif.add_edge("B", "C")
        motif.add_edge("C", "D")
        motif.add_edge("D", "A")

        host = nx.DiGraph()
        host.add_edge("A", "B")
        host.add_edge("B", "C")
        host.add_edge("C", "D")
        host.add_edge("D", "A")

        assert len(find_motifs(motif, host)) == 4

    def test_rect_count_matches_nx(self):

        host = nx.fast_gnp_random_graph(10, 0.5, directed=True)

        motif = nx.DiGraph()
        motif.add_edge("A", "B")
        motif.add_edge("B", "C")
        motif.add_edge("C", "D")
        motif.add_edge("D", "A")

        assert len(find_motifs(motif, host)) == len(
            [i for i in DiGraphMatcher(host, motif).subgraph_monomorphisms_iter()]
        )

    def test_tri_count_matches_nx(self):

        host = nx.fast_gnp_random_graph(10, 0.5, directed=True)

        motif = nx.DiGraph()
        motif.add_edge("A", "B")
        motif.add_edge("B", "C")
        motif.add_edge("C", "A")

        assert find_motifs(motif, host, count_only=True) == len(
            [i for i in DiGraphMatcher(host, motif).subgraph_monomorphisms_iter()]
        )

    def test_two_hop_count_matches_nx(self):

        host = nx.fast_gnp_random_graph(10, 0.5, directed=True)

        motif = nx.DiGraph()
        motif.add_edge("A", "B")
        motif.add_edge("B", "C")

        assert len(find_motifs(motif, host)) == len(
            [i for i in DiGraphMatcher(host, motif).subgraph_monomorphisms_iter()]
        )

    def test_high_degree_high_density_count_matches_nx(self):

        host = nx.fast_gnp_random_graph(10, 1, directed=True)

        motif = nx.DiGraph()
        motif.add_edge("A", "B")
        motif.add_edge("A", "C")
        motif.add_edge("A", "D")
        motif.add_edge("A", "E")

        assert len(find_motifs(motif, host)) == len(
            [i for i in DiGraphMatcher(host, motif).subgraph_monomorphisms_iter()]
        )

    def test_high_degree_low_density_count_matches_nx(self):

        host = nx.fast_gnp_random_graph(10, 0.3, directed=True)

        motif = nx.DiGraph()
        motif.add_edge("A", "B")
        motif.add_edge("A", "C")
        motif.add_edge("A", "D")
        motif.add_edge("A", "E")

        assert len(find_motifs(motif, host)) == len(
            [i for i in DiGraphMatcher(host, motif).subgraph_monomorphisms_iter()]
        )

    def test_falsy_node_names(self):

        motif = nx.DiGraph()
        motif.add_edge(0, 1)
        motif.add_edge(1, 2)
        motif.add_edge(2, 0)

        host = nx.DiGraph()
        host.add_edge(0, 1)
        host.add_edge(1, 2)
        host.add_edge(2, 0)

        assert len(find_motifs(motif, host)) == 3


class TestUndirectedSubgraphMatching:
    def test_subgraph_equals_graph_triangle(self):

        motif = nx.Graph()
        motif.add_edge("A", "B")
        motif.add_edge("B", "C")
        motif.add_edge("C", "A")

        host = nx.Graph()
        host.add_edge("A", "B")
        host.add_edge("B", "C")
        host.add_edge("C", "A")

        assert len(find_motifs(motif, host)) == 6

    def test_subgraph_equals_graph_rect(self):

        motif = nx.Graph()
        motif.add_edge("A", "B")
        motif.add_edge("B", "C")
        motif.add_edge("C", "D")
        motif.add_edge("D", "A")

        host = nx.Graph()
        host.add_edge("A", "B")
        host.add_edge("B", "C")
        host.add_edge("C", "D")
        host.add_edge("D", "A")

        assert len(find_motifs(motif, host)) == 8

    def test_rect_count_matches_nx(self):

        host = nx.fast_gnp_random_graph(10, 0.5, directed=False)

        motif = nx.Graph()
        motif.add_edge("A", "B")
        motif.add_edge("B", "C")
        motif.add_edge("C", "D")
        motif.add_edge("D", "A")

        assert len(find_motifs(motif, host)) == len(
            [i for i in GraphMatcher(host, motif).subgraph_monomorphisms_iter()]
        )

    def test_tri_count_matches_nx(self):

        host = nx.fast_gnp_random_graph(10, 0.5, directed=False)

        motif = nx.Graph()
        motif.add_edge("A", "B")
        motif.add_edge("B", "C")
        motif.add_edge("C", "A")

        assert len(find_motifs(motif, host)) == len(
            [i for i in GraphMatcher(host, motif).subgraph_monomorphisms_iter()]
        )

    def test_two_hop_count_matches_nx(self):

        host = nx.fast_gnp_random_graph(10, 0.5, directed=False)

        motif = nx.Graph()
        motif.add_edge("A", "B")
        motif.add_edge("B", "C")

        assert len(find_motifs(motif, host)) == len(
            [i for i in GraphMatcher(host, motif).subgraph_monomorphisms_iter()]
        )

    def test_high_degree_high_density_count_matches_nx(self):

        host = nx.fast_gnp_random_graph(10, 1, directed=False)

        motif = nx.Graph()
        motif.add_edge("A", "B")
        motif.add_edge("A", "C")
        motif.add_edge("A", "D")
        motif.add_edge("A", "E")

        assert len(find_motifs(motif, host)) == len(
            [i for i in GraphMatcher(host, motif).subgraph_monomorphisms_iter()]
        )

    def test_high_degree_low_density_count_matches_nx(self):

        host = nx.fast_gnp_random_graph(10, 0.3, directed=False)

        motif = nx.Graph()
        motif.add_edge("A", "B")
        motif.add_edge("A", "C")
        motif.add_edge("A", "D")
        motif.add_edge("A", "E")

        assert len(find_motifs(motif, host)) == len(
            [i for i in GraphMatcher(host, motif).subgraph_monomorphisms_iter()]
        )


def _random_motif():
    g = nx.graph_atlas(random.randint(7, 30))
    while len([c for c in nx.connected_components(g)]) != 1:
        g = nx.graph_atlas(random.randint(7, 30))
    return nx.relabel_nodes(g, lambda x: str(x + 1))


def _random_host(directed=False, n=20, p=0.1):
    g = nx.fast_gnp_random_graph(n, p, directed=directed)
    while (
        len(
            [
                c
                for c in (
                    nx.weakly_connected_components(g)
                    if directed
                    else nx.connected_components(g)
                )
            ]
        )
        != 1
    ):
        g = nx.fast_gnp_random_graph(n, p, directed=directed)
    return nx.relabel_nodes(g, lambda x: str(x + 1))


def _random_directed_motif():
    motif = _random_motif()
    dmotif = nx.DiGraph()
    for (u, v) in motif.edges():
        dmotif.add_edge(*random.choice([(u, v), (v, u)]))
    return dmotif


class TestRandomGraphIsomorphisms:
    @pytest.mark.parametrize(
        "host,motif",
        [(_random_host(directed=False), _random_motif()) for _ in range(5)],
    )
    def test_isomorphisms_on_undirected_random_graph(self, host, motif):
        assert find_motifs(motif, host, isomorphisms_only=True, count_only=True) == len(
            [i for i in GraphMatcher(host, motif).subgraph_isomorphisms_iter()]
        )

    @pytest.mark.parametrize(
        "host,motif",
        [(_random_host(directed=True), _random_directed_motif()) for _ in range(15)],
    )
    def test_isomorphisms_on_directed_random_graph(self, host, motif):
        assert find_motifs(
            motif, host, directed=True, isomorphisms_only=True, count_only=True
        ) == len([i for i in DiGraphMatcher(host, motif).subgraph_isomorphisms_iter()])


class TestRandomGraphMonomorphisms:
    @pytest.mark.parametrize(
        "host,motif",
        [(_random_host(directed=False), _random_motif()) for _ in range(5)],
    )
    def test_monomorphisms_on_undirected_random_graph(self, host, motif):
        assert find_motifs(motif, host, count_only=True) == len(
            [i for i in GraphMatcher(host, motif).subgraph_monomorphisms_iter()]
        )

    @pytest.mark.parametrize(
        "host,motif",
        [(_random_host(directed=True), _random_directed_motif()) for _ in range(15)],
    )
    def test_monomorphisms_on_directed_random_graph(self, host, motif):
        assert find_motifs(motif, host, directed=True, count_only=True) == len(
            [i for i in DiGraphMatcher(host, motif).subgraph_monomorphisms_iter()]
        )


class TestSubcliques:
    def test_subcliques_slow(self):
        host = nx.star_graph(30000)
        host.add_edge(6, 9)

        motif = nx.complete_graph(3)

        assert find_motifs(motif, host, count_only=True) == 6


class TestHints:
    @pytest.mark.parametrize(
        "host,motif",
        [(_random_host(directed=False), _random_motif()) for _ in range(5)],
    )
    def test_empty_hints(self, host, motif):
        assert find_motifs(motif, host, count_only=True, hints=[]) == len(
            [i for i in GraphMatcher(host, motif).subgraph_monomorphisms_iter()]
        )

    def test_broken_hints_have_no_results(self):
        host = nx.DiGraph()
        nx.add_path(host, ["A", "B", "C", "A"])
        motif = nx.DiGraph()
        nx.add_path(motif, ["A", "B", "C", "A"])
        assert (
            find_motifs(motif, host, count_only=True, hints=[{"A": "A", "B": "A"}]) == 0
        )
        assert (
            find_motifs(motif, host, count_only=True, hints=[{"A": "A", "B": "C"}]) == 0
        )

    def test_some_hints_have_values(self):
        # One mapping will fail, the other is valid:
        host = nx.DiGraph()
        nx.add_path(host, ["A", "B", "C", "A"])
        motif = nx.DiGraph()
        nx.add_path(motif, ["A", "B", "C", "A"])
        assert (
            find_motifs(
                motif,
                host,
                count_only=True,
                hints=[{"A": "A", "B": "C"}, {"A": "A", "B": "B"}],
            )
            == 1
        )

    def test_basic_hints(self):
        host = nx.DiGraph()
        nx.add_path(host, ["A", "B", "C", "A"])
        motif = nx.DiGraph()
        nx.add_path(motif, ["a", "b", "c", "a"])
        assert find_motifs(motif, host, count_only=True, hints=[{"a": "A"}]) == 1
        assert (
            find_motifs(motif, host, count_only=True, hints=[{"a": "A"}, {"b": "A"}])
            == 2
        )


class TestLimits:
    def test_zero_limit(self):
        host = nx.complete_graph(8)
        motif = nx.complete_graph(3)
        assert find_motifs(motif, host, count_only=True, limit=0) == 336

    def test_limit_one(self):
        host = nx.complete_graph(8)
        motif = nx.complete_graph(3)
        assert find_motifs(motif, host, count_only=True, limit=1) == 1

    def test_limit_eq_answer(self):
        host = nx.complete_graph(8)
        motif = nx.complete_graph(3)
        assert len(find_motifs(motif, host, limit=300)) == 300

    def test_limit_gt_answer(self):
        host = nx.complete_graph(8)
        motif = nx.complete_graph(3)
        assert find_motifs(motif, host, count_only=True, limit=338) == 336


class TestIterator:
    def test_zero_limit(self):
        host = nx.complete_graph(8)
        motif = nx.complete_graph(3)
        assert len(list(find_motifs_iter(motif, host))) == 336

    def test_can_get_next_result(self):
        host = nx.complete_graph(8)
        motif = nx.complete_graph(3)
        result = next(find_motifs_iter(motif, host))
        assert isinstance(result, dict)

    def test_fails_on_invalid_hint(self):
        host = nx.complete_graph(8)
        motif = nx.complete_graph(3)
        with pytest.raises(Exception):
            next(find_motifs_iter(motif, host, hints=[{"F": "X"}]))


class TestAttributes:
    def test_node_attributes(self):
        host = nx.DiGraph()
        nx.add_path(host, ["A", "B", "C", "A"])
        host.add_node("A", flavor="chocolate")
        host.add_node("B", flavor="coffee")
        host.add_node("C", flavor="lint")

        motif = nx.DiGraph()
        nx.add_path(motif, ["a", "b", "c", "a"])
        motif.add_node("b", flavor="chocolate")

        assert find_motifs(motif, host, count_only=True) == 1

    def test_edge_attributes(self):
        host = nx.DiGraph()
        nx.add_path(host, ["A", "B", "C", "A"])
        host.add_edge("A", "B", flavor="chocolate")
        host.add_edge("B", "C", flavor="coffee")
        host.add_edge("C", "A", flavor="lint")

        motif = nx.DiGraph()
        nx.add_path(motif, ["a", "b", "c", "a"])
        motif.add_edge("a", "b", flavor="chocolate")

        assert find_motifs(motif, host) == [{"a": "A", "b": "B", "c": "C"}]

    def test_node_and_edge_attributes(self):
        host = nx.DiGraph()
        nx.add_path(host, ["A", "B", "C", "A"])
        host.add_edge("A", "B", flavor="chocolate")
        host.add_edge("B", "C", flavor="coffee")
        host.add_edge("C", "A", flavor="lint")
        host.add_node("A", flavor="chocolate")
        host.add_node("B", flavor="coffee")
        host.add_node("C", flavor="lint")

        motif = nx.DiGraph()
        nx.add_path(motif, ["a", "b", "c", "a"])
        motif.add_edge("a", "b", flavor="coffee")
        motif.add_node("c", flavor="lint")

        assert find_motifs(motif, host) == []

    def test_attr_not_in_node(self):
        host = nx.DiGraph()
        nx.add_path(host, ["A", "B", "C", "A"])
        host.add_edge("A", "B")
        host.add_edge("B", "C")
        host.add_edge("C", "A")
        host.add_node("A")
        host.add_node("B")
        host.add_node("C")

        motif = nx.DiGraph()
        motif.add_edge("a", "b")
        motif.add_node("a", flavor="coffee")

        assert find_motifs(motif, host) == []

    def test_attr_not_in_edge(self):
        host = nx.DiGraph()
        nx.add_path(host, ["A", "B", "C", "A"])
        host.add_edge("A", "B")
        host.add_edge("B", "C")
        host.add_edge("C", "A")
        host.add_node("A")
        host.add_node("B")
        host.add_node("C")

        motif = nx.DiGraph()
        motif.add_edge("a", "b", type="delicious")

        assert find_motifs(motif, host) == []


class TestBrokenMotifFailures:
    def test_disconnected_motif(self):
        host = nx.complete_graph(8, nx.DiGraph())
        motif = nx.DiGraph()
        motif.add_node("a")
        motif.add_node("b")
        with pytest.raises(ValueError):
            find_motifs(motif, host)

    def test_empty_motif(self):
        host = nx.complete_graph(8, nx.DiGraph())
        motif = nx.DiGraph()
        with pytest.raises(ValueError):
            find_motifs(motif, host)



from collections import deque
import queue
from enum import Enum


try:
    SimpleQueue = queue.SimpleQueue
except:
    SimpleQueue = queue.Queue


class QueuePolicy(Enum):
    """
    An Enum for queue pop policy.

    DEPTHFIRST policy is identified with pushing and popping on the same side
    of the queue. BREADTHFIRST is identified with pushing and popping from the
    opposite side of the queue.

    """

    DEPTHFIRST = 1
    BREADTHFIRST = 2


class ProfilingQueue(SimpleQueue):
    """
    A simple queue to enable profiling. This queue keeps track of its size over
    time so that you can more easily debug performance.

    Note that you do not want to use this queue for production workloads, as it
    adds substantial overhead.

    """

    def __init__(self):
        """
        Create a new ProfilingQueue.

        Arguments:
            None

        Returns:
            None

        """
        super(ProfilingQueue, self).__init__()
        self._size_history = SimpleQueue()
        self._size = 0

    def put(self, *args, **kwargs):
        """
        Put a new element into the queue.

        Arguments:
            element: The element to add to the queue

        Returns:
            None

        """
        res = super(ProfilingQueue, self).put(*args, **kwargs)
        self._size += 1
        self._size_history.put(self._size)
        return res

    def get(self, *args, **kwargs):
        """
        Get a new item from the queue.

        Arguments:
            None

        Returns:
            Any: The element popped from the queue

        """
        res = super(ProfilingQueue, self).get(*args, **kwargs)
        self._size -= 1
        self._size_history.put(self._size)
        return res


class Deque:
    """
    A double-ended queue implementation.

    """

    def __init__(self, policy: QueuePolicy = QueuePolicy.DEPTHFIRST):
        """
        Create a new double-ended queue.

        Arguments:
            policy (QueuePolicy): The policy to use when adding and removing
                elements from this queue. Defaults to depth-first.

        Returns:
            None

        """
        self._dq = deque()
        if policy == QueuePolicy.DEPTHFIRST:
            self._put = self._dq.append
            self._get = self._dq.popleft
        elif policy == QueuePolicy.BREADTHFIRST:
            self._get = self._dq.pop
            self._put = self._dq.append

    def put(self, *args, **kwargs):
        """
        Put a new element into the queue.

        Arguments:
            element: The element to add to the queue

        Returns:
            None

        """
        return self._put(args[0])

    def get(self, *args, **kwargs):
        """
        Get a new item from the queue.

        Arguments:
            None

        Returns:
            Any: The element popped from the queue

        """
        return self._get()

    def empty(self):
        """
        Returns True if the queue is empty.

        Arguments:
            None

        Returns:
            bool: True if the queue has nothing more to pop

        """
        return False if self._dq else True
